<?php
declare(strict_types=1);

namespace PrometheusModule\Command;

use App\itnovum\openITCOCKPIT\Monitoring\Naemon\ExternalCommands;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Core\Configure;
use Cake\Log\Log;
use Cake\ORM\TableRegistry;
use PrometheusModule\Lib\PrometheusApi;
use PrometheusModule\Model\Table\PrometheusAlertRulesTable;

/**
 * PrometheusBridge command.
 */
class PrometheusBridgeCommand extends Command {

    /**
     * @var PrometheusApi
     */
    private $PrometheusApi;

    /**
     * @var ConsoleIo
     */
    private $io;

    /**
     * @var array
     */
    private $nonOkRules = [];

    /**
     * @var array
     */
    private $states = [];

    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/4/en/console-commands/commands.html#defining-arguments-and-options
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);

        $parser->addOptions([
            'loop'     => ['help' => __d('oitc_console', 'Run PrometheusBridge as endless loop (Daemon)'), 'boolean' => true, 'default' => false],
            'interval' => ['help' => __d('oitc_console', 'Loop interval in seconds')]
        ]);

        return $parser;
    }

    /**
     * Implement this method with your command's logic.
     *
     * @param \Cake\Console\Arguments $args The command arguments.
     * @param \Cake\Console\ConsoleIo $io The console io
     * @return null|void|int The exit code or null for success
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        $this->PrometheusApi = new PrometheusApi();
        $this->io = $io;

        $loop = $args->getOption('loop') === true;

        try {
            Configure::load('PrometheusModule.prometheus_bridge');
            $config = Configure::read('prometheus_bridge');
            $interval = (int)$config['interval'];
        } catch (\Exception $e) {
            //Config not found or not readable
            $interval = 15;
        }


        if ($args->getOption('interval') !== null) {
            $interval = $args->getOption('interval');
            if (!is_numeric($interval) || $interval < 0 || $interval > 300) {
                Log::error('Invalid interval. Interval needs to be > 0 and <= 300. Using default as fallback');
                $interval = 15;
            }
        }

        $interval = (int)$interval;

        if ($loop === true) {
            while (true) {

                try {
                    $start = microtime(true);
                    $this->loop();
                    $runtime = microtime(true) - $start;

                    $sleep = $interval - $runtime;
                    $sleep = (int)round($sleep, 0);

                    if ($sleep < 1) {
                        $sleep = 1;
                    }
                } catch (\Exception $e) {
                    Log::error($e->getMessage());
                }

                sleep($sleep);
            }
        } else {
            //Run once and exit
            $this->loop();
        }
    }

    private function loop() {
        $start = microtime(true);
        $this->io->verbose('------- Start loop -------');

        $this->states = [
            'ok'       => [
                'state_id'   => 0,
                'state_name' => __('OK')
            ],
            'warning'  => [
                'state_id'   => 1,
                'state_name' => __('Warning')
            ],
            'critical' => [
                'state_id'   => 2,
                'state_name' => __('Critical')
            ]
        ];

        //Clear alerts of the last loop
        $this->nonOkRules = [];

        $this->processAlerts();
        $this->processAllOkServices();

        $this->io->verbose('Execution time: ' . (microtime(true) - $start));
    }

    /**
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    private function processAlerts() {
        $alerts = $this->getAlerts();

        $bulkMessagesToNaemon = [];

        foreach ($alerts as $serviceId => $alertsOfService) {

            if (!empty($alertsOfService['firing']['warning']) || !empty($alertsOfService['firing']['critical'])) {
                // This Service has firing Prometheus alerts !!!

                // All alerts are from the same host/service
                $hostUuid = $alertsOfService['firing']['warning'][0]['annotations']['host_uuid'] ?? $alertsOfService['firing']['critical'][0]['annotations']['host_uuid'];
                $serviceUuid = $alertsOfService['firing']['warning'][0]['annotations']['service_uuid'] ?? $alertsOfService['firing']['critical'][0]['annotations']['service_uuid'];

                //Mark this host and service UUID as non-ok for later processing
                $key = $hostUuid . '_' . $serviceUuid;
                $this->nonOkRules[$key] = $key;

                $state = 1; // At least one metric is in a warning state
                if (!empty($alertsOfService['firing']['critical'])) {
                    $state = 2; // At least one metric is in a critical state
                }

                $warningCount = sizeof($alertsOfService['firing']['warning']);
                $criticalCount = sizeof($alertsOfService['firing']['critical']);

                $warningPendingCount = 0;
                $criticalPendingCount = 0;
                $totalPendingCount = 0;
                $criticalValues = [];
                $warningValues = [];
                if (!empty($alertsOfService['pending']['warning']) || !empty($alertsOfService['pending']['critical'])) {
                    $warningPendingCount = sizeof($alertsOfService['pending']['warning']);
                    $criticalPendingCount = sizeof($alertsOfService['pending']['critical']);
                    $totalPendingCount = $warningPendingCount + $criticalPendingCount;
                }

                $unit = '';

                // Build long output
                $longoutput = [];
                if ($criticalCount > 0) {
                    $longoutput[] = 'CRITICAL';
                    $longoutput[] = '--------';
                }
                foreach ($alertsOfService['firing']['critical'] as $cAlert) {
                    $unit = $cAlert['annotations']['unit'];
                    $labels = [];
                    foreach ($cAlert['labels'] as $labelKey => $labelValue) {
                        if ($labelKey !== 'alertname') {
                            $labels[] = sprintf('%s="%s"', $labelKey, $labelValue);
                        }
                    }
                    $longoutput[] = sprintf(
                        "%s \t %F%s",
                        implode(', ', $labels),
                        $cAlert['value'],
                        $cAlert['annotations']['unit']
                    );
                    $criticalValues[] = floatval($cAlert['value']);
                }

                if ($warningCount > 0) {
                    if ($criticalCount > 0) {
                        $longoutput[] = ''; // Add a blank line between critical and warning
                    }
                    $longoutput[] = 'WARNING';
                    $longoutput[] = '--------';
                }

                foreach ($alertsOfService['firing']['warning'] as $wAlert) {
                    $unit = $wAlert['annotations']['unit'];
                    $labels = [];
                    foreach ($wAlert['labels'] as $labelKey => $labelValue) {
                        if ($labelKey !== 'alertname') {
                            $labels[] = sprintf('%s="%s"', $labelKey, $labelValue);
                        }
                    }
                    $longoutput[] = sprintf(
                        "%s \t %F%s",
                        implode(', ', $labels),
                        $wAlert['value'],
                        $wAlert['annotations']['unit']
                    );
                    $warningValues[] = floatval($wAlert['value']);
                }

                //Build Output
                $output = '';
                if ($warningCount > 0) {
                    $output = sprintf(
                        '%s: %s item/s are warning. Warning values: [%s%s]',
                        $this->states['warning']['state_name'],
                        $warningCount,
                        implode(', ', $warningValues),
                        $unit
                    );
                }

                if ($criticalCount > 0) {
                    $output = sprintf(
                        '%s: %s item/s are critical. Critical values: [%s%s]. %s item/s are warning. Warning values: [%s%s]',
                        $this->states['critical']['state_name'],
                        $criticalCount,
                        implode(', ', $criticalValues),
                        $unit,
                        $warningCount,
                        implode(', ', $warningValues),
                        $unit
                    );
                }

                if ($totalPendingCount > 0) {
                    $output = sprintf(
                        '%s. %s items are pending.',
                        $output,
                        $totalPendingCount
                    );
                }

                $bulkMessagesToNaemon[] = [
                    'Command' => 'check_result',
                    'Data'    => [
                        'host_name'           => $hostUuid,
                        'service_description' => $serviceUuid,
                        'output'              => $output,
                        'long_output'         => (!empty($longoutput)) ? '[code]' . implode(PHP_EOL, $longoutput) . '[/code]' : '',
                        'check_type'          => 1, //https://github.com/naemon/naemon-core/blob/cec6e10cbee9478de04b4cf5af29e83d47b5cfd9/src/naemon/common.h#L330-L334
                        'return_code'         => $state,
                        'start_time'          => time(),
                        'end_time'            => time(),
                        'early_timeout'       => 0,
                        'latency'             => 0,
                        'exited_ok'           => 1
                    ]
                ];

            } // End of firing alerts if
        }

        //$datetime = \DateTime::createFromFormat('Y-m-d\TH:i:s+', $alert['activeAt'], new \DateTimeZone(date_default_timezone_get()));
        //debug($datetime);
        //debug($datetime->format('d.m.Y H:i:s'));

        if (!empty($bulkMessagesToNaemon)) {
            $this->io->verbose('Submit ' . sizeof($bulkMessagesToNaemon) . ' alerts to Naemon through Statusengine Broker');

            $ExternalCommands = new ExternalCommands();
            $ExternalCommands->toQueueBulk([
                'messages' => $bulkMessagesToNaemon
            ], 0);
        }
    }

    /**
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    private function getAlerts() {
        $alerts = $this->PrometheusApi->getPendingAndFiringAlerts();
        if (empty($alerts)) {
            $this->io->verbose('Got empty list of alerts from Prometheus');
            return [];
        }

        $worstAlerts = [];
        $this->io->verbose('Got ' . sizeof($alerts['data']['alerts']) . ' alerts from Prometheus');

        // ITC-2824 Group alerts by service_id (one PromQL can have multiple metrics)
        $combinedAlerts = [];
        foreach ($alerts['data']['alerts'] as $index => $alert) {
            $state = $alert['state']; // firing or pending
            $severity = $alert['labels']['severity']; // warning or critical
            $serviceId = $alert['annotations']['service_id'];

            if (!isset($combinedAlerts[$serviceId])) {
                $combinedAlerts[$serviceId] = [
                    'firing'  => [
                        'warning'  => [],
                        'critical' => []
                    ],
                    'pending' => [
                        'warning'  => [],
                        'critical' => []
                    ]
                ];
            }

            $combinedAlerts[$serviceId][$state][$severity][] = $alert;
        }

        return $combinedAlerts;
    }

    private function processAllOkServices() {
        /** @var PrometheusAlertRulesTable $PrometheusAlertRulesTable */
        $PrometheusAlertRulesTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusAlertRules');

        $services = $PrometheusAlertRulesTable->getAlertRulesForConfig();
        $bulkMessagesToNaemon = [];
        foreach ($services as $service) {
            $key = $service['service']['host']['uuid'] . '_' . $service['service']['uuid'];
            if (isset($this->nonOkRules[$key])) {
                //Non ok alert for current service already exists
                continue;
            }

            $result = $this->PrometheusApi->getValueByMetric($service['promql']);
            if (isset($result['status']) && $result['status'] === 'success') {

                $okValues = [];
                foreach ($result['data']['result'] as $metric) {
                    $okValues[] = floatval($metric['value'][1]);
                }

                $bulkMessagesToNaemon[] = [
                    'Command' => 'check_result',
                    'Data'    => [
                        'host_name'           => $service['service']['host']['uuid'],
                        'service_description' => $service['service']['uuid'],
                        'output'              => sprintf(
                            '%s: %s item/s are ok. Values: [%s%s]',
                            $this->states['ok']['state_name'],
                            sizeof($result['data']['result']),
                            implode(', ', $okValues),
                            $service['unit']

                        ),
                        'long_output'         => '',
                        'check_type'          => 1, //https://github.com/naemon/naemon-core/blob/cec6e10cbee9478de04b4cf5af29e83d47b5cfd9/src/naemon/common.h#L330-L334
                        'return_code'         => $this->states['ok']['state_id'],
                        'start_time'          => time(),
                        'end_time'            => time(),
                        'early_timeout'       => 0,
                        'latency'             => 0,
                        'exited_ok'           => 1
                    ]
                ];
            }

            if (sizeof($bulkMessagesToNaemon) >= 100) {
                //Do not DDoS Naemon
                $this->io->verbose('Submit ' . sizeof($bulkMessagesToNaemon) . ' Ok services to Naemon through Statusengine Broker');

                $ExternalCommands = new ExternalCommands();
                $ExternalCommands->toQueueBulk([
                    'messages' => $bulkMessagesToNaemon
                ], 0);
                $bulkMessagesToNaemon = [];
            }

        }

        if (!empty($bulkMessagesToNaemon)) {
            $this->io->verbose('Submit ' . sizeof($bulkMessagesToNaemon) . ' Ok services to Naemon through Statusengine Broker');

            $ExternalCommands = new ExternalCommands();
            $ExternalCommands->toQueueBulk([
                'messages' => $bulkMessagesToNaemon
            ], 0);
        }
    }

}
