<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace PrometheusModule\Lib;


use Cake\Filesystem\Folder;
use Cake\ORM\TableRegistry;
use PrometheusModule\Model\Table\PrometheusAlertRulesTable;
use PrometheusModule\Model\Table\PrometheusExportersTable;

class PrometheusConfigGenerator {

    public $baseDir = '/opt/openitc/prometheus';

    /**
     * @var string
     */
    private $confDir;

    public function __construct() {
        $this->confDir = $this->baseDir . '/etc';
    }

    public function createPrometheusYml() {

        /** @var PrometheusExportersTable $PrometheusExportersTable */
        $PrometheusExportersTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusExporters');

        if (!is_dir($this->confDir)) {
            mkdir($this->confDir);
        }

        if (!is_dir($this->confDir . '/targets')) {
            mkdir($this->confDir . '/targets');
        }

        $config = $this->fileHeader();

        //# Global Prometheus config
        //global:
        //  scrape_interval:     15s  # Set the scrape interval to every 15 seconds. Default is every 1 minute.
        //  evaluation_interval: 15s  # Evaluate rules every 15 seconds. The default is every 1 minute.
        //  scrape_timeout: 10s       #is set to the global default (10s).
        $config .= $this->addContent('global', 0);
        $config .= $this->addContent('scrape_interval', 2, '15s');
        $config .= $this->addContent('scrape_timeout', 2, '10s');
        $config .= $this->addContent('evaluation_interval', 2, '15s');
        $config .= PHP_EOL;
        $config .= PHP_EOL;

        //# Remote Write for VictoriaMetrics
        //remote_write:
        //  - url: http://127.0.0.1:8428/api/v1/write
        $config .= $this->addContent('remote_write', 0);
        $config .= $this->addContent('- url: http://127.0.0.1:8428/api/v1/write', 2, null, true, false);
        $config .= PHP_EOL;
        $config .= PHP_EOL;

        //# Alertmanager configuration
        //alerting:
        //  alertmanagers:
        //  - static_configs:
        //    - targets:
        //       - alertmanager:9093
        $config .= $this->addContent('#alerting', 0);
        $config .= $this->addContent('#alertmanagers', 2);
        $config .= $this->addContent('#- static_configs', 2);
        $config .= $this->addContent('#- targets', 4);
        $config .= $this->addContent('# - "127.0.0.1:9093"', 6, null, true, false);
        $config .= PHP_EOL;
        $config .= PHP_EOL;

        //# Load rules once and periodically evaluate them according to the global 'evaluation_interval'.
        //rule_files:
        // - "etc/alert-rules.yml"
        $config .= $this->addContent('rule_files', 0);
        $config .= $this->addContent(' - "etc/alert-rules.yml"', 2, null, true, false);
        $config .= PHP_EOL;
        $config .= PHP_EOL;

        //# A scrape configuration containing exactly one endpoint to scrape:
        //scrape_configs:
        //  - job_name: "nodes"
        //    scrape_interval: "15s"
        //    file_sd_configs:
        //      - files:
        //        - '/opt/openitc/prometheus/etc/targets/nodes/*.json'

        $config .= $this->addContent('scrape_configs', 0);

        //Create a single Prometheus Job for each exporter with the same scrape_interval and scrape_timeout which has an empty custom yml config
        foreach ($PrometheusExportersTable->getGroupsOfScrapeIntervalAndScrapeTimeoutForConfig() as $job) {
            $jobName = sprintf(
                '%s_%s',
                $this->replaceChars($job['scrape_interval']),
                $this->replaceChars($job['scrape_timeout'])
            );

            if (!is_dir(sprintf('%s/targets/%s', $this->confDir, $jobName))) {
                mkdir(sprintf('%s/targets/%s', $this->confDir, $jobName));
            }

            $config .= $this->addContent('- job_name', 2, sprintf('"%s"', $jobName));
            $config .= $this->addContent('scrape_interval', 4, $job['scrape_interval']);
            $config .= $this->addContent('scrape_timeout', 4, $job['scrape_timeout']);
            $config .= $this->addContent('file_sd_configs', 4);
            $config .= $this->addContent('- files', 6);
            $config .= $this->addContent(sprintf('- "%s/targets/%s/*.json"', $this->confDir, $jobName), 8, null, true, false);

            $config .= PHP_EOL;
            $config .= PHP_EOL;
        }

        //Create jobs with special own YAML config like blackbox or SNMP exporter
        foreach ($PrometheusExportersTable->getJobsWithOwnYamlConfig() as $jobWithYaml) {
            //  - job_name: 'blackbox'
            //    scrape_interval: "30s"
            //    metrics_path: "/probe"
            //    params:
            //      module: [icmp]
            //    file_sd_configs:
            //      - files:
            //        - '/etc/prometheus/targets/blackbox/*.json'
            //    relabel_configs:
            //      - source_labels: [__address__]
            //        target_label: __param_target
            //      - target_label: __address__
            //        replacement: '192.168.56.101:9115'

            $jobName = sprintf(
                '%s_%s',
                $this->replaceChars($jobWithYaml['name']),
                $jobWithYaml['id']
            );

            if (!is_dir(sprintf('%s/targets/%s', $this->confDir, $jobName))) {
                mkdir(sprintf('%s/targets/%s', $this->confDir, $jobName));
            }

            $config .= $this->addContent('- job_name', 2, sprintf('"%s"', $jobName));
            $config .= $this->addContent('scrape_interval', 4, $jobWithYaml['scrape_interval']);
            $config .= $this->addContent('scrape_timeout', 4, $jobWithYaml['scrape_timeout']);
            $config .= $this->addContent('file_sd_configs', 4);
            $config .= $this->addContent('- files', 6);
            $config .= $this->addContent(sprintf('- "%s/targets/%s/*.json"', $this->confDir, $jobName), 8, null, true, false);
            $lines = explode("\n", $jobWithYaml['yaml']);
            foreach ($lines as $line) {
                $config .= $this->addContent($line, 4, null, true, false);
            }

            $config .= PHP_EOL;
            $config .= PHP_EOL;
        }

        $file = fopen($this->baseDir . '/prometheus.yml', 'w+');
        fwrite($file, $config);
        fclose($file);
    }

    public function exportTargets() {
        if (!is_dir($this->confDir)) {
            mkdir($this->confDir);
        }

        if (!is_dir($this->confDir . '/targets')) {
            mkdir($this->confDir . '/targets');
        }

        /** @var PrometheusExportersTable $PrometheusExportersTable */
        $PrometheusExportersTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusExporters');

        $exporters = $PrometheusExportersTable->getPrometheusHostsForConfig();
        //[
        //    {
        //      "targets": ["127.0.0.1:9100"],
        //      "labels": {
        //        "uuid": "c36b8048-93ce-4385-ac19-ab5c90574b77",
        //        "instance": "prometheus",
        //        "service": "os",
        //        "__metrics_path__": "/metrics"
        //      }
        //    },
        //    {
        //      "targets": ["127.0.0.1:9104"],
        //      "labels": {
        //        "uuid": "c36b8048-93ce-4385-ac19-ab5c90574b77",
        //        "instance": "prometheus",
        //        "service": "mysql",
        //       "__metrics_path__": "/metrics"
        //      }
        //    },
        //    {
        //      "targets": ["127.0.0.1:9418"],
        //      "labels": {
        //        "uuid": "c36b8048-93ce-4385-ac19-ab5c90574b77",
        //        "instance": "prometheus",
        //        "service": "gearmand",
        //        "__metrics_path__": "/metrics"
        //      }
        //    }
        //]

        $targetGroups = [];
        foreach ($exporters as $exporter) {
            $jobName = sprintf(
                '%s_%s',
                $this->replaceChars($exporter['scrape_interval']),
                $this->replaceChars($exporter['scrape_timeout'])
            );

            if (!empty($exporter['yaml'])) {
                //Backbox or SNMP exporter with YAML config
                $jobName = sprintf(
                    '%s_%s',
                    $this->replaceChars($exporter['name']),
                    $exporter['id']
                );
            }

            if (!is_dir(sprintf('%s/targets/%s', $this->confDir, $jobName))) {
                mkdir(sprintf('%s/targets/%s', $this->confDir, $jobName));
            }

            foreach ($exporter['hosttemplates'] as $hosttemplate) {
                foreach ($hosttemplate['hosts'] as $host) {
                    //Write json config for all hosts that are using this Prometheus exporter through the host template
                    $targetAddress = $host['address']; // Just e.g. 127.0.0.1
                    if ($exporter['add_target_port'] === true) {
                        $targetAddress = sprintf('%s:%s', $host['address'], $exporter['port']); //127.0.0.1:9091
                    }

                    $jsonData = [
                        'targets' => [
                            $targetAddress
                        ],
                        'labels'  => [
                            'uuid'     => $host['uuid'],
                            'instance' => 'prometheus',
                            'service'  => $this->replaceChars($exporter['service'])
                        ]
                    ];

                    if ($exporter['metric_path'] !== '/metrics') {
                        $jsonData['labels']['__metrics_path__'] = $exporter['metric_path'];
                    }

                    $targetGroups[$jobName][] = $jsonData;
                }
            }

            foreach ($exporter['hosts'] as $host) {
                //Write json config for all hosts
                $targetAddress = $host['address']; // Just e.g. 127.0.0.1
                if ($exporter['add_target_port'] === true) {
                    $targetAddress = sprintf('%s:%s', $host['address'], $exporter['port']); //127.0.0.1:9091
                }
                $jsonData = [
                    'targets' => [
                        $targetAddress
                    ],
                    'labels'  => [
                        'uuid'     => $host['uuid'],
                        'instance' => 'prometheus',
                        'service'  => $this->replaceChars($exporter['service'])
                    ]
                ];

                if ($exporter['metric_path'] !== '/metrics') {
                    $jsonData['labels']['__metrics_path__'] = $exporter['metric_path'];
                }

                $targetGroups[$jobName][] = $jsonData;
            }
        }

        if (!empty($targetGroups)) {
            foreach ($targetGroups as $jobName => $targets) {
                $jsonFile = sprintf('%s/targets/%s/%s', $this->confDir, $jobName, 'targets_minified.json');

                $file = fopen($jsonFile, 'w+');
                fwrite($file, json_encode($targets, JSON_PRETTY_PRINT));
                fclose($file);
            }
        }
    }

    public function exportAlertRules() {
        //groups:
        //- name: default
        //  rules:
        //  - alert: ram_at_95_percent
        //    expr: node_memory_MemAvailable{job="node"} / node_memory_MemTotal{job="node"} < 0.05
        //    for: 15m
        //    labels:
        //      severity: average
        //    annotations:
        //      summary: "RAM is full"
        //      description: "RAM on {{$labels.instance}} is at 95%."
        //      graph: '1 - (node_memory_MemAvailable{instance="{{$labels.instance}}"} / node_memory_MemTotal{instance="{{$labels.instance}}"})'
        //      graph_label: "RAM Utilization [%]"
        //
        //  - alert: fs_at_80_percent
        //    expr: hrStorageUsed{job="snmp",hrStorageDescr=~"/.*"} / hrStorageSize{job="snmp",hrStorageDescr=~"/.*"} >= 0.8
        //    for: 15m
        //    labels:
        //      severity: warning
        //    annotations:
        //      summary: "File system {{$labels.hrStorageDescr}} is filling up"
        //      description: "{{$labels.hrStorageDescr}} is at 80%."
        //      graph: 'hrStorageUsed{job="snmp",instance="{{$labels.instance}}",hrStorageDescr="{{$labels.hrStorageDescr}}"} / hrStorageSize{job="snmp",instance="{{$labels.instance}}",hrStorageDescr="{{$labels.hrStorageDescr}}"} * 100'
        //      graph_label: "FS Utilization [%]"

        if (!is_dir($this->confDir)) {
            mkdir($this->confDir);
        }


        /** @var PrometheusAlertRulesTable $PrometheusAlertRulesTable */
        $PrometheusAlertRulesTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusAlertRules');

        $alertRules = $PrometheusAlertRulesTable->getAlertRulesForConfig();


        $config = $this->fileHeader();

        $config .= $this->addContent('groups', 0);
        $config .= $this->addContent('- name', 0, 'default');
        $config .= $this->addContent('rules', 2);

        foreach ($alertRules as $alertRule) {
            foreach (['warning', 'critical'] as $severity) {

                $config .= $this->addContent('- alert', 2, sprintf(
                    '%s_%s_%s',
                    $alertRule['service']['uuid'],
                    $alertRule['service']['host']['uuid'],
                    $severity
                ));

                if ($alertRule['threshold_type'] === 'scalar') {
                    // Normal threshold e.g warning = 5 and critical = 6
                    $config .= $this->addContent('expr', 4, sprintf(
                        '%s %s %s',
                        $alertRule['promql'],
                        $alertRule[$severity . '_operator'],
                        $alertRule[$severity . '_min']
                    ));
                }


                if ($alertRule['threshold_type'] === 'range_inclusive') {
                    // Range Inclusive (≥ 10 and ≤ 20 - inside the range of 10 to 20)
                    //                                  node_load15 >= 10 <= 20

                    $config .= $this->addContent('expr', 4, sprintf(
                        '%s >= %s <= %s',
                        $alertRule['promql'],
                        $alertRule[$severity . '_min'],
                        $alertRule[$severity . '_max']
                    ));
                }

                if ($alertRule['threshold_type'] === 'range_exclusive') {
                    // Range Exclusive (< 10 or > 20 - outside the range of 10 to 20)
                    //                                 node_load15 < 10 > 20

                    $config .= $this->addContent('expr', 4, sprintf(
                        '%s < %s or %s > %s',
                        $alertRule['promql'],
                        $alertRule[$severity . '_min'],
                        $alertRule['promql'],
                        $alertRule[$severity . '_max']
                    ));
                }

                $config .= $this->addContent('for', 4, $alertRule[$severity . '_longer_as']);
                $config .= $this->addContent('labels', 4);
                $config .= $this->addContent('severity', 6, $severity);
                $config .= $this->addContent('annotations', 4);
                $config .= $this->addContent('host_uuid', 6, $alertRule['service']['host']['uuid']);
                $config .= $this->addContent('host_id', 6, $alertRule['service']['host']['id']);
                $config .= $this->addContent('service_uuid', 6, $alertRule['service']['uuid']);
                $config .= $this->addContent('service_id', 6, $alertRule['service']['id']);
                $config .= $this->addContent('unit', 6, '"' . $alertRule['unit'] . '"');
            }
        }

        $config .= PHP_EOL;
        $config .= PHP_EOL;

        //Wipe the file
        fclose(fopen($this->confDir . '/alert-rules.yml', 'w+'));
        if (!empty($alertRules)) {
            $file = fopen($this->confDir . '/alert-rules.yml', 'w+');
            fwrite($file, $config);
            fclose($file);
        }

    }

    public function deleteAllConfigfiles() {
        $dir = $this->confDir . '/targets';
        if (!is_dir($dir)) {
            return;
        }

        $result = scandir($dir);
        foreach ($result as $filename) {
            if (!in_array($filename, ['.', '..'], true)) {
                if (is_dir($dir . DS . $filename)) {
                    $folder = new Folder($dir . DS . $filename);
                    $folder->delete();
                    unset($folder);
                }
            }
        }
    }

    /**
     * @param $key
     * @param int $deep
     * @param null $value
     * @param bool $newline
     * @param bool $addColonToKey
     * @return string
     */
    private function addContent($key, $deep = 1, $value = null, $newline = true, $addColonToKey = true) {
        $c = "";
        $i = 0;
        while ($i < $deep) {
            $c .= ' ';
            $i++;
        }

        if ($addColonToKey === true) {
            $key .= ':';
        }

        $c .= $key;

        if ($value !== null) {
            $c .= ' ' . $value;
        }

        if ($newline === true) {
            $c .= PHP_EOL;
        }

        return $c;
    }

    /**
     * @param string $str
     * @return string|string[]|void
     */
    private function replaceChars(string $str) {
        return preg_replace('/[^A-Za-z0-9]/', '_', $str);
    }

    /**
     * @param resource|null $file
     * @return string
     */
    private function fileHeader($file = null) {
        $header = "#########################################################################
#    DO NOT EDIT THIS FILE BY HAND -- YOUR CHANGES WILL BE OVERWRITTEN  #
#                                                                       #
#                   File generated by openITCOCKPIT                     #
#                                                                       #
#                        Created: " . date('d.m.Y H:i') . "                      #
#########################################################################

#Weblinks:
# https://prometheus.io/docs/prometheus/latest/configuration/configuration/
\n";
        if (is_resource($file)) {
            fwrite($file, $header);
        } else {
            return $header;
        }
    }

}
