<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.


declare(strict_types=1);

namespace PrometheusModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Table\HostsTable;
use App\Model\Table\ServicesTable;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\Datasource\RepositoryInterface;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;
use PrometheusModule\Lib\AlertRuleValidationRules;
use PrometheusModule\Model\Entity\PrometheusAlertRule;

/**
 * PrometheusAlertRules Model
 *
 * @property HostsTable&\Cake\ORM\Association\BelongsTo $Hosts
 * @property ServicesTable&\Cake\ORM\Association\BelongsTo $Services
 *
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule newEmptyEntity()
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule newEntity(array $data, array $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule[] newEntities(array $data, array $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule get($primaryKey, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule findOrCreate($search, ?callable $callback = null, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusAlertRule[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TimestampBehavior
 */
class PrometheusAlertRulesTable extends Table {

    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('prometheus_alert_rules');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Hosts', [
            'foreignKey' => 'host_id',
            'joinType'   => 'INNER',
            'className'  => 'Hosts',
        ]);

        $this->belongsTo('Services', [
            'foreignKey' => 'service_id',
            'joinType'   => 'INNER',
            'className'  => 'Services',
        ]);
    }

    /**
     * Bind Core2Plugin associations
     *
     * @param RepositoryInterface $coreTable
     */
    public function bindCoreAssociations(RepositoryInterface $coreTable) {
        switch ($coreTable->getAlias()) {
            case 'Hosts':
                $coreTable->hasOne('PrometheusAlertRules', [
                    'className'  => 'PrometheusModule.PrometheusAlertRules',
                    'foreignKey' => 'host_id',
                    'dependent'  => true
                ]);
                break;

            case 'Services':
                $coreTable->hasOne('PrometheusAlertRules', [
                    'className'  => 'PrometheusModule.PrometheusAlertRules',
                    'foreignKey' => 'service_id',
                    'dependent'  => true
                ]);
                break;
        }
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->requirePresence('host_id', 'create')
            ->integer('host_id')
            ->greaterThan('host_id', 0);

        $validator
            ->allowEmptyString('service_id', null, 'create')
            ->integer('service_id')
            ->greaterThan('service_id', 0);

        // Rules are used in PrometheusModule\Form\PrometheusServiceForm and \PrometheusModule\Model\Table\PrometheusAlertRulesTable;
        $validator = AlertRuleValidationRules::attachRules($validator);

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn(['host_id'], 'Hosts'));
        $rules->add($rules->existsIn(['service_id'], 'Services'));

        return $rules;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['PrometheusAlertRules.id' => $id]);
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsByServiceId($id) {
        return $this->exists(['PrometheusAlertRules.service_id' => $id]);
    }

    /**
     * @param GenericFilter $GenericFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param array $MY_RIGHTS
     * @return array
     */
    public function getAlertRulesIndex(GenericFilter $GenericFilter, $PaginateOMat = null, $MY_RIGHTS = []) {
        /** @var ServicesTable $ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');

        $where = $GenericFilter->genericFilters();

        $having = null;
        if (isset($where['servicename LIKE'])) {
            $having = [
                'servicename LIKE' => $where['servicename LIKE']
            ];
            unset($where['servicename LIKE']);
        }

        $query = $ServicesTable->find();
        $query
            ->select([
                'PrometheusAlertRules.id',
                'PrometheusAlertRules.promql',
                'PrometheusAlertRules.unit',
                'PrometheusAlertRules.host_id',
                'PrometheusAlertRules.service_id',

                'Hosts.id',
                'Hosts.name',
                'Hosts.address',
                'Hosts.uuid',
                'Hosts.container_id',

                'Services.id',
                'Services.host_id',
                'Services.uuid',
                'Services.servicetemplate_id',

                'servicename'   => $query->newExpr('IF(Services.name IS NULL, Servicetemplates.name, Services.name)'),
                'container_ids' => $query->newExpr('GROUP_CONCAT(DISTINCT HostsToContainersSharing.container_id)')
            ])
            ->innerJoin(['Hosts' => 'hosts']);
        $containerJoinConditions = [
            'HostsToContainersSharing.host_id = Hosts.id'
        ];
        if (!empty($MY_RIGHTS)) {
            $containerJoinConditions[] = [
                'HostsToContainersSharing.container_id IN ' => $MY_RIGHTS
            ];
        }
        $query->innerJoin([
            'HostsToContainersSharing' => 'hosts_to_containers'
        ], $containerJoinConditions);


        $query->innerJoinWith('Servicetemplates')
            ->innerJoinWith('PrometheusAlertRules');

        if (!empty($where)) {
            $query->where($where);
        }
        if (!empty($having)) {
            $query->having($having);
        }

        $query->group(['PrometheusAlertRules.id']);
        $query->order(
            array_merge(
                $GenericFilter->getOrderForPaginator('servicename', 'asc'),
                ['PrometheusAlertRules.id' => 'asc']
            )
        );


        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param $serviceId
     * @return PrometheusAlertRule
     * @throws RecordNotFoundException
     */
    public function getAlertRuleByServiceId($serviceId) {
        return $this->find()
            ->where(['service_id' => $serviceId])
            ->firstOrFail();
    }

    /**
     * @param $serviceId
     * @return array|\Cake\Datasource\EntityInterface
     */
    public function getAlertRuleWithServiceByServiceId($serviceId) {
        return $this->find()
            ->where(['service_id' => $serviceId])
            ->contain(['Services'])
            ->firstOrFail();
    }

    /**
     * @param $serviceId
     * @return array
     */
    public function getAlertRulesForConfig() {
        $rules = $this->find()
            ->contain([
                'Services' => function (Query $q) {
                    return $q->disableAutoFields()
                        ->select([
                            'Services.id',
                            'Services.uuid',
                            'Services.name',
                            'Services.host_id'
                        ])
                        ->contain([
                            'Hosts' => function (Query $q) {
                                return $q->disableAutoFields()
                                    ->select([
                                        'Hosts.id',
                                        'Hosts.name',
                                        'Hosts.uuid'
                                    ]);
                            }
                        ]);
                }
            ])
            ->disableHydration()
            ->all();

        if (empty($rules)) {
            return [];
        }

        return $rules->toArray();
    }

    /**
     * @param $id
     * @return array
     * @throws RecordNotFoundException
     */
    public function getAlertRuleServiceByIdForEdit($id) {
        $query = $this->find()
            ->contain([
                'Hosts'    => function (Query $q) {
                    return $q
                        ->disableAutoFields()
                        ->select([
                            'Hosts.id',
                            'Hosts.uuid',
                            'Hosts.name',
                            'Hosts.container_id'
                        ])
                        ->contain([
                            'HostsToContainersSharing'
                        ]);
                },
                'Services' => function (Query $q) {
                    return $q
                        ->disableAutoFields()
                        ->select([
                            'Services.id',
                            'Services.uuid',
                            'Services.name',
                            'Services.servicetemplate_id'
                        ])
                        ->contain([
                            'Servicetemplates' => function (Query $q) {
                                return $q
                                    ->disableAutoFields()
                                    ->select([
                                        'Servicetemplates.id',
                                        'Servicetemplates.name'
                                    ]);
                            }
                        ]);
                }
            ])
            ->where([
                'PrometheusAlertRules.service_id' => $id
            ])
            ->disableHydration();

        $alertRule = $query->firstOrFail();
        return $alertRule;
    }
}
