<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace PrometheusModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Table\ContainersTable;
use App\Model\Table\HostsTable;
use Cake\Datasource\RepositoryInterface;
use Cake\Log\Log;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Core\FileDebugger;
use itnovum\openITCOCKPIT\Core\HostConditions;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;
use PrometheusModule\Lib\ScrapeIntervals;
use Symfony\Component\Yaml\Exception\ParseException;
use Symfony\Component\Yaml\Yaml;

/**
 * PrometheusExporters Model
 *
 * @method \PrometheusModule\Model\Entity\PrometheusExporter newEmptyEntity()
 * @method \PrometheusModule\Model\Entity\PrometheusExporter newEntity(array $data, array $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter[] newEntities(array $data, array $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter get($primaryKey, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter findOrCreate($search, ?callable $callback = null, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method \PrometheusModule\Model\Entity\PrometheusExporter[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 */
class PrometheusExportersTable extends Table {

    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('prometheus_exporters');
        $this->setDisplayField('name');
        $this->setPrimaryKey('id');

        $this->belongsTo('Containers', [
            'foreignKey' => 'container_id',
            'joinType'   => 'INNER',
            'className'  => 'DistributeModule.Containers',
        ]);

        $this->belongsToMany('Hosttemplates', [
            'className'        => 'Hosttemplates',
            'foreignKey'       => 'prometheus_exporter_id',
            'targetForeignKey' => 'hosttemplate_id',
            'joinTable'        => 'hosttemplates_to_prometheus_exporters',
            'saveStrategy'     => 'replace',
            'dependent'        => true
        ]);

        $this->belongsToMany('Hosts', [
            'className'        => 'Hosts',
            'foreignKey'       => 'prometheus_exporter_id',
            'targetForeignKey' => 'host_id',
            'joinTable'        => 'hosts_to_prometheus_exporters',
            'saveStrategy'     => 'replace',
            'dependent'        => true
        ]);
    }

    /**
     * Bind Core2Plugin associations
     *
     * @param RepositoryInterface $coreTable
     */
    public function bindCoreAssociations(RepositoryInterface $coreTable) {
        switch ($coreTable->getAlias()) {
            case 'Hosttemplates':
                $coreTable->belongsToMany('PrometheusExporters', [
                    'className'        => 'PrometheusModule.PrometheusExporters',
                    'foreignKey'       => 'hosttemplate_id',
                    'targetForeignKey' => 'prometheus_exporter_id',
                    'joinTable'        => 'hosttemplates_to_prometheus_exporters',
                    'saveStrategy'     => 'replace',
                    'dependent'        => true
                ]);
                break;

            case 'Hosts':
                $coreTable->belongsToMany('PrometheusExporters', [
                    'className'        => 'PrometheusModule.PrometheusExporters',
                    'foreignKey'       => 'host_id',
                    'targetForeignKey' => 'prometheus_exporter_id',
                    'joinTable'        => 'hosts_to_prometheus_exporters',
                    'saveStrategy'     => 'replace',
                    'dependent'        => true
                ]);
                break;
        }
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $ScrapeIntervals = new ScrapeIntervals();


        $validator
            ->integer('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->scalar('name')
            ->maxLength('name', 255)
            ->requirePresence('name', 'create')
            ->notEmptyString('name');

        $validator
            ->integer('container_id')
            ->requirePresence('container_id', 'create')
            ->allowEmptyString('container_id', null, false)
            ->greaterThanOrEqual('container_id', 1);

        $validator
            ->scalar('metric_path')
            ->maxLength('metric_path', 128)
            ->notEmptyString('metric_path');

        $validator
            ->integer('port')
            ->requirePresence('port', 'create')
            ->notEmptyString('port');

        $validator
            ->boolean('add_target_port')
            ->requirePresence('port', 'create')
            ->notEmptyString('port');

        $validator
            ->scalar('service')
            ->maxLength('service', 255)
            ->requirePresence('service', 'create')
            ->notEmptyString('service');

        $validator
            ->scalar('scrape_interval')
            ->maxLength('scrape_interval', 128)
            ->requirePresence('scrape_interval', 'create')
            ->inList('scrape_interval', array_keys($ScrapeIntervals->getScrapeIntervals()), __('Scrape interval is out of bounds: [{0}]', implode(', ', array_keys($ScrapeIntervals->getScrapeIntervals()))))
            ->notEmptyString('scrape_interval');

        $validator
            ->scalar('scrape_timeout')
            ->maxLength('scrape_timeout', 128)
            ->requirePresence('scrape_timeout', 'create')
            ->inList('scrape_timeout', array_keys($ScrapeIntervals->getScrapeIntervals()), __('Scrape timeout is out of bounds: [{0}]', implode(', ', array_keys($ScrapeIntervals->getScrapeIntervals()))))
            ->notEmptyString('scrape_timeout');

        $validator
            ->scalar('yaml')
            ->add('yaml', 'custom', [
                'rule'    => [$this, 'isYaml'],
                'message' => __('This is not a valid YAML configuration.')
            ]);

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn(['container_id'], 'Containers'));

        return $rules;
    }

    public function isYaml($value, $context) {
        if (empty($context['data']['yaml'])) {
            return true;
        }

        if (isset($context['data']['yaml']) && is_string($context['data']['yaml'])) {
            if (strlen($context['data']['yaml']) > 0) {
                try {
                    $yaml = Yaml::parse($context['data']['yaml'], Yaml::PARSE_EXCEPTION_ON_INVALID_TYPE);

                    if (!is_array($yaml)) {
                        // Most likely if its not an array, its broken.
                        // Yaml::PARSE_EXCEPTION_ON_INVALID_TYPE is just a joke or so - it never works for now
                        return false;
                    }

                    return true;
                } catch (ParseException $e) {
                    Log::error('PrometheusModule: Validation YAML parse error:' . $e->getMessage());
                    return false;
                }
            }
        }

        return true;
    }

    /**
     * @param GenericFilter $GenericFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param array $MY_RIGHTS
     * @return array
     */
    public function getExportersIndex(GenericFilter $GenericFilter, $PaginateOMat = null, $MY_RIGHTS = []) {
        $query = $this->find('all')
            ->where($GenericFilter->genericFilters())
            ->order($GenericFilter->getOrderForPaginator('PrometheusExporters.name', 'asc'));

        if (!empty($MY_RIGHTS)) {
            $query->andWhere([
                'PrometheusExporters.container_id IN' => $MY_RIGHTS
            ]);
        }

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['PrometheusExporters.id' => $id]);
    }

    /**
     * @param array $containerIds
     * @param string $type
     * @param string $index
     * @return array
     */
    public function getExportersByContainerId($containerIds = [], $type = 'all', $index = 'id', $showScrapeInterval = true) {
        if (!is_array($containerIds)) {
            $containerIds = [$containerIds];
        }

        //Lookup for the tenant container of $container_id
        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');


        $tenantContainerIds = [];

        foreach ($containerIds as $containerId) {
            if ($containerId != ROOT_CONTAINER) {

                // Get container id of the tenant container
                // $container_id is may be a location, devicegroup or whatever, so we need to container id of the tenant container to load contactgroups and contacts
                $path = $ContainersTable->getPathByIdAndCacheResult($containerId, 'PrometheusExportersByContainerId');

                // Tenant contact groups are available for all users of a tenant (oITC V2 legacy)
                if (isset($path[1])) {
                    $tenantContainerIds[] = $path[1]['id'];
                }
            } else {
                $tenantContainerIds[] = ROOT_CONTAINER;
            }
        }
        $tenantContainerIds = array_unique($tenantContainerIds);
        $containerIds = array_unique(array_merge($tenantContainerIds, $containerIds));


        if (empty($containerIds)) {
            return [];
        }

        $query = $this->find()
            ->where([
                'container_id IN' => $containerIds
            ])
            ->disableHydration()
            ->all();

        $records = $query->toArray();
        if (empty($records) || is_null($records)) {
            return [];
        }

        if ($type === 'all') {
            return $records;
        }

        $list = [];
        foreach ($records as $record) {
            $name = $record['name'];

            if ($showScrapeInterval === true) {
                $name .= ' (' . $record['scrape_interval'] . ')';
            }

            $list[$record[$index]] = $name;
        }
        return $list;
    }

    /**
     * @param HostConditions $HostConditions
     * @param int|array $selected
     * @param bool $returnEmptyArrayIfMyRightsIsEmpty
     * @return array|null
     */
    public function getPrometheusHostsForAngular(HostConditions $HostConditions, $selected = [], $returnEmptyArrayIfMyRightsIsEmpty = false) {
        $MY_RIGHTS = $HostConditions->getContainerIds();
        if ($returnEmptyArrayIfMyRightsIsEmpty === true) {
            if (empty($MY_RIGHTS)) {
                //User has no permissions to edit hosts/services
                return [];
            }
        }

        $hostnameLike = null;
        $conditions = $HostConditions->getConditions();
        if (isset($conditions['Hosts.name LIKE'])) {
            $hostnameLike = $conditions['Hosts.name LIKE'];
        }

        $query = $this->find()
            ->contain([
                // Get all hosts from this host group
                'Hosts'         => function (Query $query) use ($MY_RIGHTS, $hostnameLike) {
                    $query->disableAutoFields()
                        ->select([
                            'Hosts.id',
                            'Hosts.name'
                        ]);
                    if (!empty($hostnameLike)) {
                        $query->where([
                            'Hosts.name LIKE' => $hostnameLike
                        ]);
                    }
                    if (!empty($MY_RIGHTS)) {
                        $query->innerJoin(['HostsToContainersSharing' => 'hosts_to_containers'], [
                            'HostsToContainersSharing.host_id = Hosts.id'
                        ]);
                        $query->where([
                            'HostsToContainersSharing.container_id IN' => $MY_RIGHTS
                        ]);
                    }
                    return $query;
                },
                // Get all hosts that are in this host group through the host template AND
                // which does NOT have any own host groups
                'Hosttemplates' => function (Query $query) use ($MY_RIGHTS, $hostnameLike) {
                    $query->disableAutoFields()
                        ->select([
                            'id',
                        ])
                        ->contain([
                            'Hosts' => function (Query $query) use ($MY_RIGHTS, $hostnameLike) {
                                $query->disableAutoFields()
                                    ->select([
                                        'Hosts.id',
                                        'Hosts.name',
                                        'Hosts.hosttemplate_id',
                                        'PrometheusExporters.id'
                                    ])
                                    ->leftJoinWith('PrometheusExporters')
                                    ->whereNull('PrometheusExporters.id');
                                if (!empty($hostnameLike)) {
                                    $query->where([
                                        'Hosts.name LIKE' => $hostnameLike
                                    ]);
                                }
                                if (!empty($MY_RIGHTS)) {
                                    $query->innerJoin(['HostsToContainersSharing' => 'hosts_to_containers'], [
                                        'HostsToContainersSharing.host_id = Hosts.id'
                                    ]);
                                    $query->where([
                                        'HostsToContainersSharing.container_id IN' => $MY_RIGHTS
                                    ]);
                                }
                                return $query;
                            }
                        ]);
                    return $query;
                },
            ])
            ->disableHydration()
            ->all();

        if ($query === null) {
            return [];
        }

        $query = $query->toArray();

        $hostList = [];
        foreach ($query as $record) {
            foreach ($record['hosttemplates'] as $hosttemplates) {
                foreach ($hosttemplates['hosts'] as $host) {
                    $hostList[$host['id']] = $host['name'];
                }
            }

            foreach ($record['hosts'] as $host) {
                $hostList[$host['id']] = $host['name'];
            }
        }

        asort($hostList, SORT_FLAG_CASE | SORT_NATURAL);

        return $hostList;
    }

    /**
     * Prometheus Jobs (exporters) that use the same scrape_interval and scrape_timeout can get
     * grouped together into 1 single job.
     *
     * Exporters that use a yaml config (like the blackbox exporter) requires an own job definition in prometheus.yml
     *
     * @return array
     */
    public function getGroupsOfScrapeIntervalAndScrapeTimeoutForConfig() {
        $query = $this->find()
            ->select([
                'scrape_interval',
                'scrape_timeout'
            ])
            ->where([
                'OR' => [
                    'yaml' => '',
                    'yaml IS NULL'
                ]
            ])
            ->group(
                [
                    'scrape_interval',
                    'scrape_timeout'
                ]
            )
            ->disableHydration()
            ->all();

        if (empty($query)) {
            return [];
        }

        return $query->toArray();
    }

    /**
     * Get all Jobs/Exporters that use an yaml config (like the blackbox exporter)
     *
     * @return array
     */
    public function getJobsWithOwnYamlConfig() {
        $query = $this->find()
            ->where([
                'AND' => [
                    'yaml !=' => '',
                    'yaml IS NOT NULL'
                ]
            ])
            ->disableHydration()
            ->all();

        if (empty($query)) {
            return [];
        }

        return $query->toArray();
    }

    /**
     * @return array
     */
    public function getPrometheusHostsForConfig() {
        $query = $this->find()
            ->contain([
                // Get all hosts from this Prometheus exporter
                'Hosts'         => function (Query $query) {
                    $query->disableAutoFields()
                        ->select([
                            'Hosts.id',
                            'Hosts.name',
                            'Hosts.uuid',
                            'Hosts.address'
                        ]);
                    return $query;
                },
                // Get all hosts that are using this Prometheus exporter through the host template AND
                // which does NOT have any own exporters
                'Hosttemplates' => function (Query $query) {
                    $query->disableAutoFields()
                        ->select([
                            'id',
                        ])
                        ->contain([
                            'Hosts' => function (Query $query) {
                                $query->disableAutoFields()
                                    ->select([
                                        'Hosts.id',
                                        'Hosts.name',
                                        'Hosts.uuid',
                                        'Hosts.hosttemplate_id',
                                        'Hosts.address',
                                        'PrometheusExporters.id'
                                    ])
                                    ->leftJoinWith('PrometheusExporters')
                                    ->whereNull('PrometheusExporters.id');
                                return $query;
                            }
                        ]);
                    return $query;
                },
            ])
            ->disableHydration()
            ->all();

        if ($query === null) {
            return [];
        }

        $query = $query->toArray();

        return $query;
    }
}
